#include "cloth.h"

#include <GL/glut.h>
#include <math.h>
#include <iostream.h>
#include "lights.h"
#include "control.h"
#include "vector.h"


int   stiffness  = 2;
float gravity[3] = { 0, 0, -0.04 };
float ks   = 0.9;
float kd   = 0.7;

float  cloth_vertices[CLOTH_HEIGHT*CLOTH_WIDTH*3];
float  cloth_vertices_buffer[CLOTH_HEIGHT*CLOTH_WIDTH*3];
float  cloth_velocities[CLOTH_HEIGHT*CLOTH_WIDTH*3];
float  cloth_velocities_buffer[CLOTH_HEIGHT*CLOTH_WIDTH*3];
float  cloth_wind[CLOTH_HEIGHT*CLOTH_WIDTH*3];
float  cloth_colors[CLOTH_HEIGHT*CLOTH_WIDTH*3];

triangle cloth_triangles[NUM_TRIANGLES];
GLuint   cloth_indices[NUM_INDICES];

bool   anchor   = true;
bool   sphere   = false;
bool   wind     = false;
bool   internal = false;

float scenter[3] = { 0, 0, 4 };
float sradius = 2.0;

float windvec[3] = { 0, 0.01, 0 };


void swapClothBuffers ()
{
  int index;

  for (int i = 0; i < CLOTH_WIDTH; i++)
    for (int j = 0; j < CLOTH_HEIGHT; j++)
    {
      index = (j*CLOTH_WIDTH + i) * 3;
      cloth_vertices[index]   = cloth_vertices_buffer[index];
      cloth_vertices[++index] = cloth_vertices_buffer[index];
      cloth_vertices[++index] = cloth_vertices_buffer[index];

      index -= 2;
      cloth_velocities[index]   = cloth_velocities_buffer[index];
      cloth_velocities[++index] = cloth_velocities_buffer[index];
      cloth_velocities[++index] = cloth_velocities_buffer[index];
    }
}


void initCloth ()
{
  for (int i = 0; i < CLOTH_WIDTH; i++)
    for (int j = 0; j < CLOTH_HEIGHT; j++)
    {
      cloth_vertices[(j*CLOTH_WIDTH+i)*3 + 0] = (i - ((CLOTH_WIDTH-1) / 2.0))
                                              * GRID_SIZE;
      cloth_vertices[(j*CLOTH_WIDTH+i)*3 + 1] = (j - ((CLOTH_HEIGHT-1) / 2.0))
                                              * GRID_SIZE;
      cloth_vertices[(j*CLOTH_WIDTH+i)*3 + 2] = 15.0;

      cloth_velocities[(j*CLOTH_WIDTH+i)*3 + 0] = 0.0;
      cloth_velocities[(j*CLOTH_WIDTH+i)*3 + 1] = 0.0;
      cloth_velocities[(j*CLOTH_WIDTH+i)*3 + 2] = 0.0;

      cloth_colors[(j*CLOTH_WIDTH+i)*3 + 0] = i / (float) CLOTH_WIDTH;
      cloth_colors[(j*CLOTH_WIDTH+i)*3 + 1] = j / (float) CLOTH_HEIGHT;
      cloth_colors[(j*CLOTH_WIDTH+i)*3 + 2] = 1 - (j / (float) CLOTH_HEIGHT);
    }

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);

  glVertexPointer(3, GL_FLOAT, 0, cloth_vertices);
  glColorPointer(3, GL_FLOAT, 0, cloth_colors);

  int current = 0;

  for (int i = 0; i < CLOTH_WIDTH; i++)
    for (int j = 0; j < CLOTH_HEIGHT; j++)
    {
      if ((i < CLOTH_WIDTH-1) && (j < CLOTH_HEIGHT-1))
      {
          cloth_indices[current*3 + 0] = j*CLOTH_WIDTH + i;
          cloth_indices[current*3 + 1] = j*CLOTH_WIDTH + (i+1);
          cloth_indices[current*3 + 2] = (j+1)*CLOTH_WIDTH + i;

          cloth_triangles[current].v[0] = (j*CLOTH_WIDTH + i)*3;
          cloth_triangles[current].v[1] = (j*CLOTH_WIDTH + (i+1))*3;
          cloth_triangles[current].v[2] = ((j+1)*CLOTH_WIDTH + i)*3;

          current++;
      }

      if ((i > 0) && (j > 0))
      {
          cloth_indices[current*3 + 0] = j*CLOTH_WIDTH + i;
          cloth_indices[current*3 + 1] = (j-1)*CLOTH_WIDTH + i;
          cloth_indices[current*3 + 2] = j*CLOTH_WIDTH + (i-1);

          cloth_triangles[current].v[0] = (j*CLOTH_WIDTH + i)*3;
          cloth_triangles[current].v[1] = ((j-1)*CLOTH_WIDTH + i)*3;
          cloth_triangles[current].v[2] = (j*CLOTH_WIDTH + (i-1))*3;

          current++;
      }      
    }
}


void drawCloth ()
{
  glDisable(GL_LIGHTING);
  glDrawElements(GL_TRIANGLES, NUM_INDICES, 
                 GL_UNSIGNED_INT, cloth_indices);
  glEnable(GL_LIGHTING);
}


bool anchored (int i, int j)
{
  return anchor && 
         (j == 0);
}


void calculateNormals ()
{
  float a[3], b[3];

  for (int i = 0; i < NUM_TRIANGLES; i++)
  {
    for (int j = 0; j < 3; j++)
    {
      a[j] = cloth_vertices[cloth_triangles[i].v[1] + j] 
           - cloth_vertices[cloth_triangles[i].v[0] + j];
      b[j] = cloth_vertices[cloth_triangles[i].v[2] + j] 
           - cloth_vertices[cloth_triangles[i].v[0] + j];
    }

    vectorCross(cloth_triangles[i].n, a, b);
    vectorMag(&(cloth_triangles[i].m), cloth_triangles[i].n);
  }
}


void calculateWind ()
{
  float windF[3], dot;

  for (int i = 0; i < CLOTH_WIDTH*CLOTH_HEIGHT*3; i++)
    cloth_wind[i] = 0.0;

  for (int i = 0; i < NUM_TRIANGLES; i++)
  {
    vectorDot(&dot, cloth_triangles[i].n, windvec);
    vectorMult(windF, dot / cloth_triangles[i].m, cloth_triangles[i].n);

    for (int j = 0; j < 3; j++)
      vectorAddV(&(cloth_wind[cloth_triangles[i].v[j]]), 1,
                 &(cloth_wind[cloth_triangles[i].v[j]]), 1,
                 windF);
  }
}


void updateCloth ()
{
  int index, neighbor;
  float springV[3], springL, restL;
  float dampI[3], dampM;
  float forceS[3], forceV[3], forceM;

  if (wind || internal)
    calculateNormals();

  if (wind)
    calculateWind();

  for (int i = 0; i < CLOTH_WIDTH; i++)
    for (int j = 0; j < CLOTH_HEIGHT; j++)
      if (anchored(i,j))
      {
        index = (j*CLOTH_WIDTH + i) * 3;
        vectorCopy(&(cloth_vertices_buffer[index]),
                   &(cloth_vertices[index]));
      }
      else
      {
        index = (j*CLOTH_WIDTH + i) * 3;

        vectorSet(forceV, 0.0);
        vectorAddV(forceV, 1, forceV, 1, gravity);

        for (int u = -stiffness; u <= stiffness; u++)
          for (int v = -stiffness; v <= stiffness; v++)
            if ((i+u >= 0) && (i+u < CLOTH_WIDTH) &&
                (j+v >= 0) && (j+v < CLOTH_HEIGHT) &&
                ((u != 0) || (v != 0)))
            {
                neighbor = ((j+v)*CLOTH_WIDTH + (i+u)) * 3;

                vectorAddV(springV, -1, &(cloth_vertices[neighbor]),
                           1, &(cloth_vertices[index]));

                // spring contribution
                vectorMag(&springL, springV);
                if (springL == 0)
                    cerr << "zero-length spring!" << endl;
                else
                    vectorMult(springV, 1/springL, springV);
                restL  = sqrt(sqr(u*GRID_SIZE) + sqr(v*GRID_SIZE));
                forceM = - ks * (springL - restL);

                // damping contribution
                vectorAddV(dampI, -1, &(cloth_velocities[neighbor]),
                           1, &(cloth_velocities[index]));
                vectorDot(&dampM, dampI, springV);
                forceM -= kd * dampM;

                vectorMult(forceS, forceM, springV);
                if (wind)
                  vectorAddV(forceS, 1, forceS, 1, &(cloth_wind[index]));
                vectorAddV(forceV, 1, forceV, 1, forceS);
            }

        vectorAddV(&(cloth_velocities_buffer[index]), 
                   1, &(cloth_velocities[index]),
                   dtime, forceV);

        vectorAddV(&(cloth_vertices_buffer[index]), 
                   1, &(cloth_vertices[index]),
                   dtime, &(cloth_velocities_buffer[index]));

        // collisions
        floorCollision(index);
        if (internal)
            internalCollision(index);
        if (sphere)
            sphereCollision(index);
      }

  swapClothBuffers();
}


float intersect (const float p1_[3], const float p2_[3], triangle* t)
{
  float nd, a;
  float d[3], w[3], p1[3];

  vectorAddV(d, 1, p2_, -1, p1_);
  vectorDot(&nd, t->n, d);

  if (nd == 0)
      return -1.0;

  vectorAddV(p1, 1, p1_, -1, &(cloth_vertices[t->v[0]]));
  vectorDot(&a, t->n, p1);
  a = -a / nd;

  vectorAddV(w, 1, p1_, a, d);
  vectorAddV(w, 1, w, -1, &(cloth_vertices[t->v[0]]));



  return -1.0;
}


void internalCollision (int index)
{
  for (int i = 0; i < NUM_TRIANGLES; i++)
    if (intersect(&(cloth_vertices[index]),
                  &(cloth_vertices_buffer[index]), 
                  &(cloth_triangles[i])) >= 0)
    {
      vectorCopy(&(cloth_vertices_buffer[index]), &(cloth_vertices[index]));
    }
}


void floorCollision (int index)
{
  if (cloth_vertices_buffer[index + 2] < 0.01)
  {
    cloth_vertices_buffer[index + 2]   = 0.01;
    cloth_velocities_buffer[index + 2] = 0;

    // friction
    vectorMult(&(cloth_velocities_buffer[index]), 0.1,
               &(cloth_velocities_buffer[index]));

    float m;
    vectorMag(&m, &(cloth_velocities_buffer[index]));
    if (m < 0.05)
      vectorSet(&(cloth_velocities_buffer[index]), 0);
  }
}


void sphereCollision (int index)
{
  float v[3], d;

  vectorAddV(v, -1, scenter, 1, &(cloth_vertices_buffer[index]));
  vectorMag(&d, v);

  if (d < sradius + 0.1)
  {
    float vr[3], dot;

    vectorMult(v, 1/d, v);
    vectorMult(vr, (sradius + 0.1), v);
    vectorAddV(vr, 1, vr, 1, scenter);
    vectorCopy(&(cloth_vertices_buffer[index]), vr);

    vectorDot(&dot, v, &(cloth_velocities_buffer[index]));
    vectorMult(v, dot, v);
    vectorAddV(&(cloth_velocities_buffer[index]), 1,
               &(cloth_velocities_buffer[index]), -1, v);

    // friction
    vectorMult(&(cloth_velocities_buffer[index]), 0.1,
               &(cloth_velocities_buffer[index]));

    float m;
    vectorMag(&m, &(cloth_velocities_buffer[index]));
    if (m < 0.02)
      vectorSet(&(cloth_velocities_buffer[index]), 0);
  }
}
